#include <iostream>

#include "TFile.h"
#include "TTree.h"
#include "TIterator.h"
#include "TStopwatch.h"

#include "Event_CStyleArray.h"
#include "Event_VarArray.h"
#include "Event_Vector.h"
#include "StatsDisplay.h"

template <class EventTemplate>
void writeEvents(Int_t nEvents, Int_t nPulses)
{
	std::cout << "Writing " << nEvents << " events ... ";
	TFile outF("tree.root", "RECREATE");
	TTree outT("tree", "A Benchmark tree with Events");

	EventTemplate *event = new EventTemplate();
	outT.Branch("events", &event);

	TStopwatch timer;
	timer.Start();

	for (Int_t ev = 0; ev < nEvents; ev++) {
		event->Build(ev);

		for (Int_t p = 0; p < nPulses; p++) {
			event->AddPulse();
		}

		outT.Fill();
		event->Clear();
	}
	outF.Write();

	timer.Stop();
	std::cout << "Done." << std::endl;

	Double_t rtime = timer.RealTime();
	Double_t ctime = timer.CpuTime();
	gStatsDisplay->fGFileSize->SetPoint(gStatsDisplay->fGFileSize->GetN(), nEvents, Double_t(outF.GetSize())/1024/1024);
	gStatsDisplay->fGWriteCPUSpeed->SetPoint(gStatsDisplay->fGWriteCPUSpeed->GetN(), nEvents, Double_t(nEvents/ctime));
	gStatsDisplay->fGWriteRealSpeed->SetPoint(gStatsDisplay->fGWriteRealSpeed->GetN(), nEvents, Double_t(nEvents/rtime));
}

template <class EventTemplate, class PulseTemplate>
void readEvents()
{
	TFile *inF = new TFile("tree.root");
	TTree *inT = (TTree*)inF->Get("tree");

	EventTemplate *event = new EventTemplate();
	inT->SetBranchAddress("events", &event);

	// We'll just see how many pulses are positive and negative
	Long64_t nPositivePulses = 0, nNegativePulses = 0;

	Long64_t nevents = inT->GetEntries();
	std::cout << "Reading " << nevents << " events ... ";

	TStopwatch timer;
	timer.Start();
	Long64_t nbytes = 0;

	for (Long64_t ev = 0; ev < nevents; ev++) {
		nbytes += inT->GetEntry(ev);

		TIter next(event->GetPulses());
		PulseTemplate *pulse;
		while ( pulse = (PulseTemplate*)next() ) {
			if (pulse->GetPolarity() == true) {
				nPositivePulses++;
			} else {
				nNegativePulses++;
			}
		}
		event->Clear();
	}

	timer.Stop();
	std::cout << "Done." << std::endl;
	inF->Close();

	Double_t rtime = timer.RealTime();
	Double_t ctime = timer.CpuTime();
	gStatsDisplay->fGReadCPUSpeed->SetPoint(gStatsDisplay->fGReadCPUSpeed->GetN(), nevents, nevents/ctime);
	gStatsDisplay->fGReadRealSpeed->SetPoint(gStatsDisplay->fGReadRealSpeed->GetN(), nevents, nevents/rtime);
	gStatsDisplay->fGEventSize->SetPoint(gStatsDisplay->fGEventSize->GetN(), nevents, Double_t(nbytes)/1024/nevents);
}

int main()
{
	int arr_eventSizes[] = {100, 500, 1000, 2000, 5000, 10000};
	int npulses = 100;
	std::vector<int> eventSizes(arr_eventSizes, arr_eventSizes + sizeof(arr_eventSizes)/sizeof(int));

	// C Style Array
	std::cout << " == C-Style Array Benchmarks ==" << std::endl;
	gStatsDisplay->ResetGraphs();
	for (std::vector<int>::iterator it = eventSizes.begin(); it != eventSizes.end(); it++) {
		writeEvents<Event_CStyleArray>(*it, npulses);
		readEvents<Event_CStyleArray, Pulse_CStyleArray>();
	}
	gStatsDisplay->AddAllGraphs("C-Style Array", kRed-4);

	// ROOT Style Variable Array
	std::cout << " == Variable Length Array Benchmarks ==" << std::endl;
	gStatsDisplay->ResetGraphs();
	for (std::vector<int>::iterator it = eventSizes.begin(); it != eventSizes.end(); it++) {
		writeEvents<Event_VarArray>(*it, npulses);
		readEvents<Event_VarArray, Pulse_VarArray>();
	}
	gStatsDisplay->AddAllGraphs("Variable Array", kAzure+7);

	// std::vector<>
	std::cout << " == std::vector Benchmarks ==" << std::endl;
	gStatsDisplay->ResetGraphs();
	for (std::vector<int>::iterator it = eventSizes.begin(); it != eventSizes.end(); it++) {
		writeEvents<Event_Vector>(*it, npulses);
		readEvents<Event_Vector, Pulse_Vector>();
	}
	gStatsDisplay->AddAllGraphs("Vector", kSpring-1);

	gStatsDisplay->Draw();
	gStatsDisplay->Save("statsfile.root");

	return 0;
}
